@@ -34,6 +34,10 @@
 (require 'dash-functional)
 (require 'goto-addr)
 (require 'markdown-mode)
+(require 'cl-lib)
+
+(when (featurep 'xwidget-internal)
+  (require 'xwidget))
 
 (defgroup lsp-ui-doc nil
   "Display informations of the current line."
@@ -58,13 +62,20 @@
   :type 'boolean
   :group 'lsp-ui-doc)
 
-(defcustom lsp-ui-doc-position 'top
+(defcustom lsp-ui-doc-position 'at-point
   "Where to display the doc."
   :type '(choice (const :tag "Top" top)
                  (const :tag "Bottom" bottom)
                  (const :tag "At point" at-point))
   :group 'lsp-ui-doc)
 
+(defcustom lsp-ui-doc-alignment 'frame
+  "How to align the doc.
+This only takes effect when `lsp-ui-doc-position' is 'top or 'bottom."
+  :type '(choice (const :tag "Frame" frame)
+                 (const :tag "Window" window))
+  :group 'lsp-ui-doc)
+
 (defcustom lsp-ui-doc-border "white"
   "Border color of the frame."
   :type 'color
@@ -86,6 +97,23 @@ Child frames requires GNU/Emacs version >= 26 and graphical frames."
   :type 'boolean
   :group 'lsp-ui-doc)
 
+(defcustom lsp-ui-doc-use-webkit nil
+  "Whether to display documentation in a WebKit widget in a child-frame.
+This requires GNU/Emacs version >= 26 and built with the `--with-xwidgets`
+option."
+  :type 'boolean
+  :group 'lsp-ui-doc)
+
+(defcustom lsp-ui-doc-delay 0.2
+  "Number of seconds before showing the doc."
+  :type 'number
+  :group 'lsp-ui-doc)
+
+(defcustom lsp-ui-doc-winum-ignore t
+  "Whether to ignore lsp-ui-doc buffers in winum."
+  :type 'boolean
+  :group 'lsp-ui-doc)
+
 (defface lsp-ui-doc-background
   '((((background light)) :background "#b3b3b3")
     (t :background "#272A36"))
@@ -106,7 +134,6 @@ Only the `background' is used in this face."
 
 (defvar lsp-ui-doc-frame-parameters
   '((left . -1)
-    (no-accept-focus . t)
     (no-focus-on-map . t)
     (min-width  . 0)
     (width  . 0)
@@ -125,7 +152,6 @@ Only the `background' is used in this face."
     (visibility . nil)
     (mouse-wheel-frame . nil)
     (no-other-frame . t)
-    (cursor-type . nil)
     (inhibit-double-buffering . t)
     (drag-internal-border . t)
     (no-special-glyphs . t)
@@ -138,15 +164,16 @@ The function takes a string as parameter and should return a string.
 If this variable is nil (the default), the documentation will be rendered
 as markdown.")
 
-(defvar lsp-ui-doc-custom-markup-modes
-  '((rust-mode "no_run" "rust,no_run" "rust,ignore" "rust,should_panic"))
-  "Mode to uses with markdown code blocks.
-They are added to `markdown-code-lang-modes'")
-
 (defvar lsp-ui-doc-frame-hook nil
   "Hooks run on child-frame creation.
 The functions receive 2 parameters: the frame and its window.")
 
+(defvar lsp-ui-doc-webkit-client-path
+  (concat "file://"
+          (file-name-directory (or load-file-name buffer-file-name))
+          "lsp-ui-doc.html")
+  "Path to the page loaded when a WebKit widget is created.")
+
 ;; Avoid warning with emacs < 26
 (declare-function display-buffer-in-child-frame "window.el")
 
@@ -157,11 +184,15 @@ Because some variables are buffer local.")
 (defvar-local lsp-ui-doc--inline-ov nil
   "Overlay used to display the documentation in the buffer.")
 
+(defvar-local lsp-ui-doc--bounds nil)
+(defvar-local lsp-ui-doc--timer nil)
+
+(defconst lsp-ui-doc--buffer-prefix " *lsp-ui-doc-")
+
 (defmacro lsp-ui-doc--with-buffer (&rest body)
   "Execute BODY in the lsp-ui-doc buffer."
   `(let ((parent-vars (list :buffer (current-buffer)
-                            :window (get-buffer-window)
-                            :workspace-root (lsp-workspace-root))))
+                            :window (get-buffer-window))))
      (with-current-buffer (get-buffer-create (lsp-ui-doc--make-buffer-name))
        (setq lsp-ui-doc--parent-vars parent-vars)
        (prog1 (let ((buffer-read-only nil))
@@ -183,7 +214,7 @@ Because some variables are buffer local.")
 
 (defun lsp-ui-doc--make-buffer-name ()
   "Construct the buffer name, it should be unique for each frame."
-  (concat " *lsp-ui-doc-"
+  (concat lsp-ui-doc--buffer-prefix
           (or (frame-parameter nil 'window-id)
               (frame-parameter nil 'name))
           "*"))
@@ -192,60 +223,39 @@ Because some variables are buffer local.")
 ;; Markdown 2.3.
 (defvar markdown-fontify-code-block-default-mode)
 
-(defun lsp-ui-doc--setup-markdown (mode)
-  "Setup the ‘markdown-mode’ in the frame.
-MODE is the mode used in the parent frame."
-  (make-local-variable 'markdown-code-lang-modes)
-  (dolist (mark (alist-get mode lsp-ui-doc-custom-markup-modes))
-    (add-to-list 'markdown-code-lang-modes (cons mark mode)))
-  (setq-local markdown-fontify-code-blocks-natively t)
-  (setq-local markdown-fontify-code-block-default-mode mode)
-  (setq-local markdown-hide-markup t))
-
 (defun lsp-ui-doc--inline-wrapped-line (string)
   "Wraps a line of text for inline display."
-  (cond ((string-empty-p string) "")
-        ((< (length string) lsp-ui-doc-max-width) string)
-        (t (concat (substring string 0 (- lsp-ui-doc-max-width 1))
-                   "\\\n"
-                   (lsp-ui-doc--inline-wrapped-line (substring string (- lsp-ui-doc-max-width 1)))))))
+  (let ((doc-max-width (lsp-ui-doc--inline-window-width)))
+    (cond ((string-empty-p string) "")
+          ((< (length string) doc-max-width) string)
+          (t (concat (substring string 0 (- doc-max-width 4))
+                     "\n"
+                     (string-trim-left
+                      (lsp-ui-doc--inline-wrapped-line
+                       (substring string (- doc-max-width 4)))))))))
 
 (defun lsp-ui-doc--inline-formatted-string (string)
   "Formats STRING for inline rendering."
   (mapconcat (lambda (line)
                (lsp-ui-doc--inline-wrapped-line (string-trim-right line)))
-             (split-string string "\n")
+             (split-string string "[\n\v\f\r]")
              "\n"))
 
-(defun lsp-ui-doc--extract-marked-string (marked-string)
+(defun lsp-ui-doc--extract-marked-string (marked-string &optional language)
   "Render the MARKED-STRING."
   (string-trim-right
    (let* ((string (if (stringp marked-string)
                       marked-string
                     (gethash "value" marked-string)))
           (with-lang (hash-table-p marked-string))
-          (language (and with-lang (gethash "language" marked-string)))
-          (render-fn (if with-lang (lsp-get-renderer language)
-                       (and (functionp lsp-ui-doc-render-function)
-                            lsp-ui-doc-render-function)))
-          (mode major-mode))
-     (if render-fn
-         (funcall render-fn string)
-       (with-temp-buffer
-         (if (lsp-ui-doc--inline-p)
-             (insert (lsp-ui-doc--inline-formatted-string string))
-           (insert string))
-
-         (delay-mode-hooks
-           (let ((inhibit-message t))
-             (funcall (cond ((and with-lang (string= "text" language)) 'text-mode)
-                            ((fboundp 'gfm-view-mode) 'gfm-view-mode)
-                            (t 'markdown-mode))))
-           (when (derived-mode-p 'markdown-mode)
-             (lsp-ui-doc--setup-markdown mode))
-           (ignore-errors
-             (font-lock-ensure)))
-         (buffer-string))))))
+          (language (or (and with-lang (or (gethash "language" marked-string) (gethash "kind" marked-string)))
+                        language)))
+     (cond
+      (lsp-ui-doc-use-webkit
+       (if (and language (not (string= "text" language)))
+           (format "```%s\n%s\n```" language string)
+         string))
+      (t (lsp--render-element (lsp-ui-doc--inline-formatted-string string)))))))
 
 (defun lsp-ui-doc--filter-marked-string (list-marked-string)
   (let ((groups (--separate (and (hash-table-p it)
@@ -261,26 +271,81 @@ CONTENTS can be differents type of values:
 MarkedString | MarkedString[] | MarkupContent (as defined in the LSP).
 We don't extract the string that `lps-line' is already displaying."
   (cond
-   ((stringp contents) contents)
+   ((stringp contents) (lsp-ui-doc--extract-marked-string contents)) ;; MarkedString
    ((sequencep contents) ;; MarkedString[]
     (mapconcat 'lsp-ui-doc--extract-marked-string
-               (lsp-ui-doc--filter-marked-string contents)
+               (lsp-ui-doc--filter-marked-string (seq-filter #'identity contents))
                "\n\n"
                ;; (propertize "\n\n" 'face '(:height 0.4))
                ))
    ;; when we get markdown contents, render using emacs gfm-view-mode / markdown-mode
-   ((string= (gethash "kind" contents) "markdown") (lsp-ui-doc--extract-marked-string contents))
-   ((gethash "kind" contents) (gethash "value" contents)) ;; MarkupContent
+   ((string= (gethash "kind" contents) "markdown") ;; Markdown MarkupContent
+    (lsp-ui-doc--extract-marked-string (gethash "value" contents) "markdown"))
+   ((gethash "kind" contents) (gethash "value" contents)) ;; Plaintext MarkupContent
    ((gethash "language" contents) ;; MarkedString
-    (lsp-ui-doc--extract-marked-string contents))))
+    (lsp-ui-doc--extract-marked-string (gethash "value" contents)
+                                       (gethash "language" contents)))))
+
+(defun lsp-ui-doc--webkit-run-xwidget ()
+  "Launch embedded WebKit instance."
+  (lsp-ui-doc--with-buffer
+   (let ((inhibit-read-only t))
+     (insert " ")
+     (goto-char 1)
+     (let ((id (make-xwidget
+                'webkit
+                nil
+                1
+                1
+                nil
+                (buffer-name))))
+       (set-xwidget-query-on-exit-flag id nil)
+       (put-text-property (point) (+ 1 (point))
+                          'display (list 'xwidget ':xwidget id))
+       (xwidget-webkit-mode)
+       (xwidget-webkit-goto-uri (xwidget-at 1)
+                                lsp-ui-doc-webkit-client-path)
+       (lsp-ui-doc--webkit-set-background)
+       (lsp-ui-doc--webkit-set-foreground)))))
+
+(defun lsp-ui-doc--webkit-set-background ()
+  "Set background color of the WebKit widget."
+  (lsp-ui-doc--webkit-execute-script
+   (format "document.body.style.background = '%s';"
+           "#fdfdfd"
+           ;; (face-attribute 'lsp-ui-doc-background :background)
+           )))
+
+(defun lsp-ui-doc--webkit-set-foreground ()
+  "Set foreground color of the WebKit widget."
+  (lsp-ui-doc--webkit-execute-script
+   (format "document.body.style.color = '%s';"
+           (face-attribute 'default :foreground))))
+
+(defun lsp-ui-doc--webkit-get-xwidget ()
+  "Return Xwidget instance."
+  (lsp-ui-doc--with-buffer
+    (xwidget-at 1)))
+
+(defun lsp-ui-doc--webkit-execute-script (script &optional fn)
+  "Execute SCRIPT in embedded Xwidget and run optional callback FN."
+  (when-let* ((xw (lsp-ui-doc--webkit-get-xwidget)))
+    (xwidget-webkit-execute-script xw script fn)))
+
+(defun lsp-ui-doc--webkit-execute-script-rv (script)
+  "Execute SCRIPT in embedded Xwidget synchronously."
+  (when-let* ((xw (lsp-ui-doc--webkit-get-xwidget)))
+    (xwidget-webkit-execute-script-rv xw script)))
 
 (defun lsp-ui-doc--hide-frame ()
   "Hide the frame."
+  (setq lsp-ui-doc--bounds nil)
   (when (overlayp lsp-ui-doc--inline-ov)
     (delete-overlay lsp-ui-doc--inline-ov))
   (when (lsp-ui-doc--get-frame)
-    (lsp-ui-doc--with-buffer
-     (erase-buffer))
+    (unless lsp-ui-doc-use-webkit
+      (lsp-ui-doc--with-buffer
+       (erase-buffer)))
     (make-frame-invisible (lsp-ui-doc--get-frame))))
 
 (defun lsp-ui-doc--buffer-width ()
@@ -298,9 +363,11 @@ We don't extract the string that `lps-line' is already displaying."
 
 (defun lsp-ui-doc--line-height (&optional line)
   "Return the pos-y of the LINE on screen, in pixel."
-  (nth 2 (or (window-line-height line)
-             (and (redisplay t)
-                  (window-line-height line)))))
+  (or
+   (nth 2 (or (window-line-height line)
+              (and (redisplay t)
+                   (window-line-height line))))
+   0))
 
 (defun lsp-ui-doc--sideline-pos-y ()
   (-> (when (bound-and-true-p lsp-ui-sideline--occupied-lines)
@@ -308,56 +375,68 @@ We don't extract the string that `lps-line' is already displaying."
       (line-number-at-pos)
       (lsp-ui-doc--line-height)))
 
+(defun lsp-ui-doc--webkit-resize-callback (size)
+  (let ((offset-width (round (aref size 0)))
+        (offset-height (round (aref size 1))))
+    (xwidget-resize (lsp-ui-doc--webkit-get-xwidget) offset-width offset-height))
+  (lsp-ui-doc--move-frame (lsp-ui-doc--get-frame)))
+
 (defun lsp-ui-doc--resize-buffer ()
   "If the buffer's width is larger than the current frame, resize it."
-  (let* ((frame-width (frame-width))
-         (fill-column (min lsp-ui-doc-max-width (- frame-width 5))))
-    (when (> (lsp-ui-doc--buffer-width) (min lsp-ui-doc-max-width frame-width))
-      (lsp-ui-doc--with-buffer
-       (fill-region (point-min) (point-max))))))
+  (if lsp-ui-doc-use-webkit
+       (lsp-ui-doc--webkit-execute-script
+        "[document.querySelector('#lsp-ui-webkit').offsetWidth, document.querySelector('#lsp-ui-webkit').offsetHeight];"
+       'lsp-ui-doc--webkit-resize-callback)
 
-(defun lsp-ui-doc--next-to-side-window-p nil
-  "Return non-nil if the window on the left is a side window."
-  (let* ((win (window-at 0 0))
-         (left (window-left (selected-window))))
-    (and (not (eq win (selected-window)))
-         (or (not left) (eq win left))
-         (eq (window-parameter win 'window-side) 'left))))
+    (let* ((frame-width (frame-width))
+           (fill-column (min lsp-ui-doc-max-width (- frame-width 5))))
+      (when (> (lsp-ui-doc--buffer-width) (min lsp-ui-doc-max-width frame-width))
+        (lsp-ui-doc--with-buffer
+         (fill-region (point-min) (point-max)))))))
 
-(defun lsp-ui-doc--mv-at-point (frame height start-x start-y)
-  "Move the FRAME at point.
+(defun lsp-ui-doc--mv-at-point (frame width height start-x start-y)
+  "Move FRAME to be where the point is.
+WIDTH is the child frame width.
 HEIGHT is the child frame height.
 START-X is the position x of the current window.
-START-Y is the position y of the current window."
+START-Y is the position y of the current window.
+The algorithm prefers to position FRAME just above the
+symbol at point, to not obstruct the view of the code that follows.
+If there's no space above in the current window, it places
+FRAME just below the symbol at point."
   (-let* (((x . y) (--> (bounds-of-thing-at-point 'symbol)
-                        (nth 2 (posn-at-point (car it)))))
-          (mode-line-y (lsp-ui-doc--line-height 'mode-line))
+                        (posn-x-y (posn-at-point (car it)))))
+          (frame-relative-symbol-x (+ start-x x))
+          (frame-relative-symbol-y (+ start-y y))
           (char-height (frame-char-height))
-          (y (or (and (> y (/ mode-line-y 2))
-                      (<= (- mode-line-y y) (+ char-height height))
-                      (> (- y height) 0)
-                      (- y height))
-                 (+ y char-height))))
-    (set-frame-position frame (+ x start-x) (+ y start-y))))
+          ;; Make sure the frame is positioned horizontally such that
+          ;; it does not go beyond the frame boundaries.
+          (frame-x (or (and (<= (frame-outer-width) (+ frame-relative-symbol-x width))
+                            (- x (- (+ frame-relative-symbol-x width)
+                                    (frame-outer-width))))
+                       x))
+          (frame-y (or (and (<= height frame-relative-symbol-y)
+                            (- y height))
+                       (+ y char-height))))
+    (set-frame-position frame (+ start-x frame-x) (+ start-y frame-y))))
 
 (defun lsp-ui-doc--move-frame (frame)
   "Place our FRAME on screen."
-  (lsp-ui-doc--resize-buffer)
-  (-let* (((left top _right _bottom) (window-edges nil nil nil t))
+  (-let* (((left top right _bottom) (window-edges nil nil nil t))
           (window (frame-root-window frame))
           ((width . height) (window-text-pixel-size window nil nil 10000 10000 t))
           (width (+ width (* (frame-char-width frame) 1))) ;; margins
           (char-h (frame-char-height))
           (height (min (- (* lsp-ui-doc-max-height char-h) (/ char-h 2)) height))
+          (frame-right (pcase lsp-ui-doc-alignment
+                         ('frame (frame-pixel-width))
+                         ('window right)))
           (frame-resize-pixelwise t))
     (set-frame-size frame width height t)
     (if (eq lsp-ui-doc-position 'at-point)
-        (lsp-ui-doc--mv-at-point frame height left top)
+        (lsp-ui-doc--mv-at-point frame width height left top)
       (set-frame-position frame
-                          (if (and (>= left (+ width 10 (frame-char-width)))
-                                   (not (lsp-ui-doc--next-to-side-window-p)))
-                              10
-                            (- (frame-pixel-width) width 10 (frame-char-width)))
+                          (max (- frame-right width 10 (frame-char-width)) 10)
                           (pcase lsp-ui-doc-position
                             ('top (+ top 10))
                             ('bottom (- (lsp-ui-doc--line-height 'mode-line)
@@ -396,13 +475,20 @@ FN is the function to call on click."
 (defun lsp-ui-doc--render-buffer (string symbol)
   "Set the buffer with STRING."
   (lsp-ui-doc--with-buffer
-   (erase-buffer)
-   (let ((inline-p (lsp-ui-doc--inline-p)))
-     (insert (concat (unless inline-p (propertize "\n" 'face '(:height 0.2)))
-                     (-> (replace-regexp-in-string "`\\([\n]+\\)" "" string)
-                         (string-trim-right))
-                     (unless inline-p (propertize "\n\n" 'face '(:height 0.3))))))
-   (lsp-ui-doc--make-clickable-link)
+   (if lsp-ui-doc-use-webkit
+       (progn
+         (lsp-ui-doc--webkit-execute-script
+          (format
+           "renderMarkdown('%s', '%s');"
+           symbol
+           (url-hexify-string string))
+          'lsp-ui-doc--webkit-resize-callback))
+     (erase-buffer)
+     (let ((inline-p (lsp-ui-doc--inline-p)))
+       (insert (concat (unless inline-p (propertize "\n" 'face '(:height 0.2)))
+                       (s-trim string)
+                       (unless inline-p (propertize "\n\n" 'face '(:height 0.3))))))
+     (lsp-ui-doc--make-clickable-link))
    (setq-local face-remapping-alist `((header-line lsp-ui-doc-header)))
    (setq-local window-min-height 1)
    (setq header-line-format (when lsp-ui-doc-header (concat " " symbol))
@@ -515,6 +601,8 @@ HEIGHT is the documentation number of lines."
 
 (defun lsp-ui-doc--display (symbol string)
   "Display the documentation."
+  (when (and lsp-ui-doc-use-webkit (not (featurep 'xwidget-internal)))
+    (setq lsp-ui-doc-use-webkit nil))
   (if (or (null string) (string-empty-p string))
       (lsp-ui-doc--hide-frame)
     (lsp-ui-doc--render-buffer string symbol)
@@ -522,7 +610,9 @@ HEIGHT is the documentation number of lines."
         (lsp-ui-doc--inline)
       (unless (lsp-ui-doc--get-frame)
         (lsp-ui-doc--set-frame (lsp-ui-doc--make-frame)))
-      (lsp-ui-doc--move-frame (lsp-ui-doc--get-frame))
+      (unless lsp-ui-doc-use-webkit
+        (lsp-ui-doc--resize-buffer)
+        (lsp-ui-doc--move-frame (lsp-ui-doc--get-frame)))
       (unless (frame-visible-p (lsp-ui-doc--get-frame))
         (make-frame-visible (lsp-ui-doc--get-frame))))))
 
@@ -534,7 +624,8 @@ HEIGHT is the documentation number of lines."
          (name-buffer (lsp-ui-doc--make-buffer-name))
          (buffer (get-buffer name-buffer))
          (params (append lsp-ui-doc-frame-parameters
-                         `((default-minibuffer-frame . ,(selected-frame))
+                         `((name . "")
+                           (default-minibuffer-frame . ,(selected-frame))
                            (minibuffer . ,(minibuffer-window))
                            (left-fringe . ,(frame-char-width))
                            (background-color . ,(face-background 'lsp-ui-doc-background nil t)))))
@@ -542,23 +633,94 @@ HEIGHT is the documentation number of lines."
                   buffer
                   `((child-frame-parameters . ,params))))
          (frame (window-frame window)))
+    (with-current-buffer buffer
+      (lsp-ui-doc-frame-mode 1))
     (set-frame-parameter nil 'lsp-ui-doc-buffer buffer)
     (set-window-dedicated-p window t)
     (redirect-frame-focus frame (frame-parent frame))
     (set-face-background 'internal-border lsp-ui-doc-border frame)
+    (set-face-background 'fringe nil frame)
     (run-hook-with-args 'lsp-ui-doc-frame-hook frame window)
+    (when lsp-ui-doc-use-webkit
+      (define-key (current-global-map) [xwidget-event]
+        (lambda ()
+          (interactive)
+
+          (let ((xwidget-event-type (nth 1 last-input-event)))
+            ;; (when (eq xwidget-event-type 'load-changed)
+            ;;   (lsp-ui-doc--move-frame (lsp-ui-doc--get-frame)))
+
+            (when (eq xwidget-event-type 'javascript-callback)
+              (let ((proc (nth 3 last-input-event))
+                    (arg (nth 4 last-input-event)))
+                (funcall proc arg))))))
+      (lsp-ui-doc--webkit-run-xwidget))
     frame))
 
+(defun lsp-ui-doc--make-request nil
+  "Request the documentation to the LS."
+  (when (and (not (eq this-command 'lsp-ui-doc-hide))
+             (not (bound-and-true-p lsp-ui-peek-mode))
+             (lsp--capability "hoverProvider"))
+    (-if-let (bounds (or (and (symbol-at-point) (bounds-of-thing-at-point 'symbol))
+                         (and (looking-at "[[:graph:]]") (cons (point) (1+ (point))))))
+        (unless (equal lsp-ui-doc--bounds bounds)
+          (lsp--send-request-async
+           (lsp--make-request "textDocument/hover" (lsp--text-document-position-params))
+           (lambda (hover) (lsp-ui-doc--callback hover bounds (current-buffer)))))
+      (lsp-ui-doc--hide-frame))))
+
+(defun lsp-ui-doc--callback (hover bounds buffer)
+  "Process the received documentation.
+HOVER is the doc returned by the LS.
+BOUNDS are points of the symbol that have been requested.
+BUFFER is the buffer where the request has been made."
+  (if (and hover
+           (>= (point) (car bounds)) (<= (point) (cdr bounds))
+           (eq buffer (current-buffer)))
+      (progn
+        (setq lsp-ui-doc--bounds bounds)
+        (and lsp-ui-doc--timer (cancel-timer lsp-ui-doc--timer))
+        (setq lsp-ui-doc--timer
+              (run-with-idle-timer
+               lsp-ui-doc-delay nil
+               (lambda nil
+                 (lsp-ui-doc--display
+                  (thing-at-point 'symbol t)
+                  (-some->> (gethash "contents" hover)
+                    lsp-ui-doc--extract
+                    (replace-regexp-in-string "\r" "")))))))
+    (lsp-ui-doc--hide-frame)))
+
 (defun lsp-ui-doc--delete-frame ()
   "Delete the child frame if it exists."
   (-when-let (frame (lsp-ui-doc--get-frame))
     (delete-frame frame)
     (lsp-ui-doc--set-frame nil)))
 
-(defadvice select-window (after lsp-ui-doc--select-window activate)
-  "Delete the child frame if window changes."
-  (unless (equal (ad-get-arg 0) (selected-window))
-    (lsp-ui-doc--hide-frame)))
+(defun lsp-ui-doc--visible-p ()
+  "Return whether the LSP UI doc is visible"
+  (or (overlayp lsp-ui-doc--inline-ov)
+      (and (lsp-ui-doc--get-frame)
+           (frame-visible-p (lsp-ui-doc--get-frame)))))
+
+(defun lsp-ui--hide-doc-frame-on-window-change (fun window &optional no-record)
+  "Delete the child frame if currently selected window changes.
+Does nothing if the newly-selected window is the same window as
+before, or if the new window is the minibuffer."
+  (let ((initial-window (selected-window)))
+    (prog1 (funcall fun window no-record)
+      (unless no-record
+        (when (lsp-ui-doc--visible-p)
+          (let* ((current-window (selected-window))
+                 (doc-buffer (get-buffer (lsp-ui-doc--make-buffer-name))))
+            (unless (or (window-minibuffer-p current-window)
+                        (equal current-window initial-window)
+                        (and doc-buffer
+                             (equal (window-buffer initial-window) doc-buffer)))
+              (lsp-ui-doc--hide-frame))))))))
+
+(advice-add #'select-window :around #'lsp-ui--hide-doc-frame-on-window-change)
 
 (advice-add 'load-theme :before (lambda (&rest _) (lsp-ui-doc--delete-frame)))
 (add-hook 'window-configuration-change-hook #'lsp-ui-doc--hide-frame)
@@ -570,15 +732,6 @@ HEIGHT is the documentation number of lines."
             (and (buffer-live-p it) it)
             (kill-buffer it)))
 
-(defun lsp-ui-doc--on-hover (hover)
-  "Handler for `lsp-on-hover-hook'.
-HOVER is the returned signature information."
-  (--if-let (-some->> hover (gethash "contents"))
-      (lsp-ui-doc--display (thing-at-point 'symbol t)
-                           (lsp-ui-doc--extract it))
-    (eldoc-message nil)
-    (lsp-ui-doc--hide-frame)))
-
 (define-minor-mode lsp-ui-doc-mode
   "Minor mode for showing hover information in child frame."
   :init-value nil
@@ -595,11 +748,11 @@ HOVER is the returned signature information."
         ;; ‘frameset-filter-alist’ for explanation.
         (cl-callf copy-tree frameset-filter-alist)
         (push '(lsp-ui-doc-frame . :never) frameset-filter-alist)))
-
-    (add-hook 'lsp-on-hover-hook 'lsp-ui-doc--on-hover nil t)
+    (add-hook 'post-command-hook 'lsp-ui-doc--make-request nil t)
     (add-hook 'delete-frame-functions 'lsp-ui-doc--on-delete nil t))
    (t
-    (remove-hook 'lsp-on-hover-hook 'lsp-ui-doc--on-hover t)
+    (lsp-ui-doc-hide)
+    (remove-hook 'post-command-hook 'lsp-ui-doc--make-request t)
     (remove-hook 'delete-frame-functions 'lsp-ui-doc--on-delete t))))
 
 (defun lsp-ui-doc-enable (enable)
@@ -610,12 +763,53 @@ It is supposed to be called from `lsp-ui--toggle'"
 (defun lsp-ui-doc-show ()
   "Trigger display hover information popup."
   (interactive)
-  (lsp-ui-doc--on-hover (lsp-request "textDocument/hover" (lsp--text-document-position-params))))
+  (lsp-ui-doc--callback (lsp-request "textDocument/hover" (lsp--text-document-position-params))
+                        (or (bounds-of-thing-at-point 'symbol) (cons (point) (1+ (point))))
+                        (current-buffer)))
 
 (defun lsp-ui-doc-hide ()
   "Hide hover information popup."
   (interactive)
   (lsp-ui-doc--hide-frame))
 
+(defun lsp-ui-doc--glance-hide-frame ()
+  "Hook to hide hover information popup for lsp-ui-doc-glance."
+  (cl-letf (((symbol-function 'frame?) (lambda (frame)
+                                         (and frame
+                                              (frame-visible-p frame)
+                                              (not (frame-focus-state frame))))))
+    (when (or (overlayp lsp-ui-doc--inline-ov)
+              (frame? (lsp-ui-doc--get-frame)))
+      (lsp-ui-doc-hide)
+      (remove-hook 'post-command-hook 'lsp-ui-doc--glance-hide-frame))))
+
+(defun lsp-ui-doc-glance ()
+  "Trigger display hover information popup and hide it on next typing."
+  (interactive)
+  (lsp-ui-doc-show)
+  (add-hook 'post-command-hook 'lsp-ui-doc--glance-hide-frame))
+
+(define-minor-mode lsp-ui-doc-frame-mode
+  "Marker mode to add additional key bind for lsp-ui-doc-frame."
+  :init-value nil
+  :lighter ""
+  :group lsp-ui-doc
+  :keymap `(([?q] . lsp-ui-doc-unfocus-frame)))
+
+(defun lsp-ui-doc-focus-frame ()
+  "Focus into lsp-ui-doc-frame."
+  (interactive)
+  (let ((frame (lsp-ui-doc--get-frame)))
+    (when (and frame (frame-visible-p frame))
+      (lsp-ui-doc--with-buffer
+       (setq cursor-type t))
+      (select-frame-set-input-focus frame))))
+
+(defun lsp-ui-doc-unfocus-frame ()
+  "Unfocus from lsp-ui-doc-frame."
+  (interactive)
+  (when-let ((frame (frame-parent (lsp-ui-doc--get-frame))))
+    (select-frame-set-input-focus frame)))
+
 (provide 'lsp-ui-doc)
 ;;; lsp-ui-doc.el ends here
