@@ -25,16 +25,18 @@
 
 ;;; Commentary:
 ;;
-;; Utility to show informations of the current line
+;; Utility to show information for the current line
 
 ;;; Code:
 
 (require 'lsp-mode)
 (require 'flycheck nil 'noerror)
 (require 'dash)
+(require 'seq)
+(require 'subr-x)
 
 (defgroup lsp-ui-sideline nil
-  "Display informations of the current line."
+  "Display information for the current line."
   :group 'tools
   :group 'convenience
   :group 'lsp-ui
@@ -47,7 +49,7 @@
   :group 'lsp-ui)
 
 (defcustom lsp-ui-sideline-ignore-duplicate nil
-  "Control to ignore duplicates when there is a same symbol with the same contents."
+  "Ignore duplicates when there is a same symbol with the same contents."
   :type 'boolean
   :group 'lsp-ui-sideline)
 
@@ -56,7 +58,7 @@
   :type 'boolean
   :group 'lsp-ui-sideline)
 
-(defcustom lsp-ui-sideline-show-hover t
+(defcustom lsp-ui-sideline-show-hover nil
   "Whether to show hover messages in sideline."
   :type 'boolean
   :group 'lsp-ui-sideline)
@@ -71,7 +73,7 @@
   :type 'boolean
   :group 'lsp-ui-sideline)
 
-(defcustom lsp-ui-sideline-update-mode 'line
+(defcustom lsp-ui-sideline-update-mode 'point
   "Define the mode for updating sideline information.
 
 When set to `line' the information will be updated when user
@@ -86,10 +88,28 @@ when user changes current point."
   :type 'number
   :group 'lsp-ui-sideline)
 
+(defcustom lsp-ui-sideline-diagnostic-max-lines 20
+  "Maximum number of lines to show of diagnostics in sideline."
+  :type 'integer
+  :group 'lsp-ui-sideline)
+
+(defcustom lsp-ui-sideline-diagnostic-max-line-length 100
+  "Maximum line length of diagnostics in sideline."
+  :type 'integer
+  :group 'lsp-ui-sideline)
+
+(defcustom lsp-ui-sideline-actions-kind-regex "quickfix.*\\|refactor.*"
+  "Regex for the code actions kinds to show in the sideline."
+  :type 'string
+  :group 'lsp-ui-sideline)
+
 (defvar lsp-ui-sideline-code-actions-prefix ""
   "Prefix to insert before the code action title.
 This can be used to insert, for example, an unicode character: 💡")
 
+(defvar-local lsp-ui-sideline--requests nil
+  "Pending requests sent by `lsp-ui-sideline'.")
+
 (defvar-local lsp-ui-sideline--ovs nil
   "Overlays used by `lsp-ui-sideline'.")
 
@@ -137,6 +157,13 @@ It is used to know when the window has changed of width.")
 This face have a low priority over the others."
   :group 'lsp-ui-sideline)
 
+(defun lsp-ui-sideline--first-line-p (pos)
+  "Return non-nil if POS is on the first line."
+  (save-excursion
+    (goto-char 1)
+    (forward-line)
+    (> (point) pos)))
+
 (defun lsp-ui-sideline--calc-space (win-width str-len index)
   "Calcul whether there is enough space on line.
 If there is enough space, it returns the point of the last
@@ -152,24 +179,29 @@ INDEX is the line number (relative to the current line)."
         (when (>= (- win-width (current-column)) str-len)
           eol)))))
 
-(defun lsp-ui-sideline--find-line (str-len &optional up)
+(defun lsp-ui-sideline--find-line (str-len bol eol &optional up offset)
   "Find a line where the string can be inserted.
 It loops on the nexts lines to find enough space.
 Returns the point of the last character on the line.
 
-WIN-WIDTH is the window width.
 STR-LEN is the string size.
-if UP is non-nil, it loops on the previous lines.."
+BOL & EOL are beginning and ending of the user point line.
+if UP is non-nil, it loops on the previous lines.
+if OFFSET is non-nil, it starts search OFFSET lines from user point line."
   (let ((win-width (lsp-ui-sideline--window-width))
-        (index 1) pos)
+        (index (if (null offset) 1 offset))
+        pos)
     (while (and (null pos) (<= (abs index) 30))
       (setq index (if up (1- index) (1+ index)))
       (setq pos (lsp-ui-sideline--calc-space win-width str-len index)))
-    (when pos (push pos lsp-ui-sideline--occupied-lines))
-    (if (or (equal pos (point-min))
-            (and up (null pos)))
-        (lsp-ui-sideline--find-line str-len)
-      pos)))
+    (if (and up (or (null pos) (and (lsp-ui-sideline--first-line-p pos)
+                                    ;; line-end-position returns a wrong value when its
+                                    ;; argument lead to a line < 0, so we need to use this trick
+                                    (-any-p 'lsp-ui-sideline--first-line-p lsp-ui-sideline--occupied-lines))))
+        (lsp-ui-sideline--find-line str-len bol eol nil offset)
+      (and pos (or (> pos eol) (< pos bol))
+           (push pos lsp-ui-sideline--occupied-lines)
+           (list pos index)))))
 
 (defun lsp-ui-sideline--delete-ov ()
   "Delete overlays."
@@ -189,24 +221,16 @@ function signature)."
       (seq-find (lambda (it) (and (hash-table-p it)
                                   (lsp-get-renderer (gethash "language" it))))
                 contents))
-     ((gethash "kind" contents) (gethash "value" contents)) ;; MarkupContent
-     ((gethash "language" contents) ;; MarkedString
-      (and (lsp-get-renderer (gethash "language" contents))
-           (gethash "value" contents))))))
+     ((gethash "kind" contents) contents) ;; MarkupContent
+     ((gethash "language" contents) contents))))
 
 (defun lsp-ui-sideline--format-info (marked-string)
   "Format MARKED-STRING.
 If the string has a language, we fontify it with the function provided
 by `lsp-mode'.
 MARKED-STRING is the string returned by `lsp-ui-sideline--extract-info'."
-  (when marked-string
-    (when (hash-table-p marked-string)
-      (let* ((language (gethash "language" marked-string))
-             (value (gethash "value" marked-string))
-             (renderer (lsp-get-renderer language)))
-        (setq marked-string (if (and (functionp renderer) value)
-                                (funcall renderer value)
-                              value))))
+  (when (and marked-string (hash-table-p marked-string))
+    (setq marked-string (lsp--render-element marked-string))
     (add-face-text-property 0 (length marked-string) 'lsp-ui-sideline-symbol-info nil marked-string)
     (add-face-text-property 0 (length marked-string) 'default t marked-string)
     (replace-regexp-in-string "[\n\t ]+" " " marked-string)))
@@ -216,9 +240,9 @@ MARKED-STRING is the string returned by `lsp-ui-sideline--extract-info'."
      (if (display-graphic-p) 1 2)))
 
 (defun lsp-ui-sideline--make-display-string (info symbol current)
-  "Make final string to display on buffer.
+  "Make final string to display in buffer.
 INFO is the information to display.
-SYMBOL is the symbol associated to the info.
+SYMBOL is the symbol associated with the info.
 CURRENT is non-nil when the point is on the symbol."
   (let* ((face (if current 'lsp-ui-sideline-current-symbol 'lsp-ui-sideline-symbol))
          (str (if lsp-ui-sideline-show-symbol
@@ -244,6 +268,7 @@ CURRENT is non-nil when the point is on the symbol."
               (or (equal (car fringe-mode) 0)
                   (equal (cdr fringe-mode) 0))
               1)
+         (and (boundp 'fringe-mode) (equal fringe-mode 0) 1)
          0)
      (if (bound-and-true-p display-line-numbers-mode)
          (+ 2 (line-number-display-width))
@@ -258,18 +283,19 @@ CURRENT is non-nil when the point is on the symbol."
   (- (min (window-text-width) (window-body-width))
      (lsp-ui-sideline--margin-width)))
 
-(defun lsp-ui-sideline--push-info (symbol tag bounds info)
+(defun lsp-ui-sideline--push-info (symbol tag bounds info bol eol)
   (when (and (= tag (lsp-ui-sideline--calculate-tag))
              (not (lsp-ui-sideline--stop-p)))
-    (let* ((info (concat (thread-first (gethash "contents" info)
+    (let* ((info (concat (-some->> (gethash "contents" info)
                            lsp-ui-sideline--extract-info
-                           lsp-ui-sideline--format-info)))
+                           lsp-ui-sideline--format-info
+                           (replace-regexp-in-string "\r" ""))))
            (current (and (>= (point) (car bounds)) (<= (point) (cdr bounds)))))
       (when (and (> (length info) 0)
                  (lsp-ui-sideline--check-duplicate symbol info))
         (let* ((final-string (lsp-ui-sideline--make-display-string info symbol current))
-               (pos-ov (lsp-ui-sideline--find-line (length final-string)))
-               (ov (when pos-ov (make-overlay pos-ov pos-ov))))
+               (pos-ov (lsp-ui-sideline--find-line (length final-string) bol eol))
+               (ov (when pos-ov (make-overlay (car pos-ov) (car pos-ov)))))
           (when pos-ov
             (overlay-put ov 'info info)
             (overlay-put ov 'symbol symbol)
@@ -277,6 +303,7 @@ CURRENT is non-nil when the point is on the symbol."
             (overlay-put ov 'current current)
             (overlay-put ov 'after-string final-string)
             (overlay-put ov 'window (get-buffer-window))
+            (overlay-put ov 'kind 'info)
             (push ov lsp-ui-sideline--ovs)))))))
 
 (defun lsp-ui-sideline--toggle-current (ov current)
@@ -299,29 +326,46 @@ CURRENT is non-nil when the point is on the symbol."
         (when (overlay-get ov 'current)
           (lsp-ui-sideline--toggle-current ov nil))))))
 
-(defun lsp-ui-sideline--diagnostics ()
+(defun lsp-ui-sideline--split-long-lines (lines)
+  "Fill LINES so that they are not longer than `lsp-ui-sideline-diagnostic-max-line-length' characters."
+  (cl-mapcan (lambda (line)
+               (if (< (length line) lsp-ui-sideline-diagnostic-max-line-length)
+                   (list line)
+                 (with-temp-buffer
+                   (let ((fill-column lsp-ui-sideline-diagnostic-max-line-length))
+                     (insert line)
+                     (fill-region (point-min) (point-max))
+                     (split-string (buffer-string) "\n")))))
+             lines))
+
+(defun lsp-ui-sideline--diagnostics (bol eol)
   "Show diagnostics on the current line."
-  (let ((bol (line-beginning-position))
-        (eol (line-end-position)))
-    (when (bound-and-true-p flycheck-mode)
+  (when (bound-and-true-p flycheck-mode)
       (dolist (e (flycheck-overlay-errors-in bol (1+ eol)))
-        (let* ((message (--> (flycheck-error-format-message-and-id e)
-                             (car (split-string it "\n"))
-                             (replace-regexp-in-string "[\n\t ]+" " " it)))
-               (len (length message))
-               (level (flycheck-error-level e))
-               (face (if (eq level 'info) 'success level))
-               (margin (lsp-ui-sideline--margin-width))
-               (message (progn (add-face-text-property 0 len 'lsp-ui-sideline-global nil message)
-                               (add-face-text-property 0 len face nil message)
-                               message))
-               (string (concat (propertize " " 'display `(space :align-to (- right-fringe ,(lsp-ui-sideline--align len margin))))
-                               message))
-               (pos-ov (lsp-ui-sideline--find-line len t))
-               (ov (and pos-ov (make-overlay pos-ov pos-ov))))
-          (when pos-ov
-            (overlay-put ov 'after-string string)
-            (push ov lsp-ui-sideline--ovs)))))))
+        (let* ((lines (--> (flycheck-error-format-message-and-id e)
+                           (split-string it "\n")
+                           (lsp-ui-sideline--split-long-lines it)))
+               (display-lines (butlast lines (- (length lines) lsp-ui-sideline-diagnostic-max-lines)))
+               (offset 1))
+          (dolist (line display-lines)
+            (let* ((message (string-trim (replace-regexp-in-string "[\t ]+" " " line)))
+                   (len (length message))
+                   (level (flycheck-error-level e))
+                   (face (if (eq level 'info) 'success level))
+                   (margin (lsp-ui-sideline--margin-width))
+                   (message (progn (add-face-text-property 0 len 'lsp-ui-sideline-global nil message)
+                                   (add-face-text-property 0 len face nil message)
+                                   message))
+                   (string (concat (propertize " " 'display `(space :align-to (- right-fringe ,(lsp-ui-sideline--align len margin))))
+                                   message))
+                   (pos-ov (lsp-ui-sideline--find-line len bol eol nil offset))
+                   (ov (and pos-ov (make-overlay (car pos-ov) (car pos-ov))))
+                   )
+              (when pos-ov
+                (setq offset (car (cdr pos-ov)))
+                (overlay-put ov 'after-string string)
+                (overlay-put ov 'kind 'diagnotics)
+                (push ov lsp-ui-sideline--ovs))))))))
 
 (defvar-local lsp-ui-sideline--code-actions nil)
 
@@ -330,17 +374,21 @@ CURRENT is non-nil when the point is on the symbol."
   (interactive)
   (unless lsp-ui-sideline--code-actions
     (user-error "No code actions on the current line"))
-  (let* ((actions lsp-ui-sideline--code-actions)
-         (title (completing-read "Apply: " (--map (gethash "title" it) actions)
-                                 nil t))
-         (action (--first (equal (gethash "title" it) title) actions)))
-    (unless action
-      (error "Fail to apply action"))
-    (lsp-execute-code-action action)))
+  (lsp-execute-code-action (lsp--select-action lsp-ui-sideline--code-actions)))
 
-(defun lsp-ui-sideline--code-actions (actions)
+(defun lsp-ui-sideline--code-actions (actions bol eol)
   "Show code ACTIONS."
+  (when lsp-ui-sideline-actions-kind-regex
+    (setq actions (seq-filter (-lambda ((&hash "kind"))
+                                (or (not kind)
+                                    (s-match lsp-ui-sideline-actions-kind-regex kind)))
+                             actions)))
   (setq lsp-ui-sideline--code-actions actions)
+  (dolist (ov lsp-ui-sideline--ovs)
+    (when (eq (overlay-get ov 'kind) 'actions)
+      (setq lsp-ui-sideline--occupied-lines
+            (delq (overlay-get ov 'position) lsp-ui-sideline--occupied-lines))
+      (delete-overlay ov)))
   (seq-doseq (action actions)
     (-let* ((title (->> (gethash "title" action)
                         (replace-regexp-in-string "[\n\t ]+" " ")
@@ -358,22 +406,24 @@ CURRENT is non-nil when the point is on the symbol."
                           title))
             (string (concat (propertize " " 'display `(space :align-to (- right-fringe ,(lsp-ui-sideline--align len margin))))
                             title))
-            (pos-ov (lsp-ui-sideline--find-line (1+ (length title)) t))
-            (ov (and pos-ov (make-overlay pos-ov pos-ov))))
+            (pos-ov (lsp-ui-sideline--find-line (1+ (length title)) bol eol t))
+            (ov (and pos-ov (make-overlay (car pos-ov) (car pos-ov)))))
       (when pos-ov
         (overlay-put ov 'after-string string)
+        (overlay-put ov 'kind 'actions)
+        (overlay-put ov 'position (car pos-ov))
         (push ov lsp-ui-sideline--ovs)))))
 
 (defun lsp-ui-sideline--calculate-tag()
-  "Calculate the tag used to determinie whether to update sideline information."
+  "Calculate the tag used to determine whether to update sideline information."
   (if (equal lsp-ui-sideline-update-mode 'line)
       (line-number-at-pos)
     (point)))
 
 (defun lsp-ui-sideline--run ()
-  "Show informations (flycheck + lsp).
-It loops on the symbols of the current line and request information
-to the language server."
+  "Show information (flycheck + lsp).
+It loops on the symbols of the current line and requests information
+from the language server."
   (lsp-ui-sideline--delete-ov)
   (when buffer-file-name
     (let ((eol (line-end-position))
@@ -386,7 +436,7 @@ to the language server."
               lsp-ui-sideline--tag tag
               lsp-ui-sideline--last-width (window-text-width))
         (when lsp-ui-sideline-show-diagnostics
-          (lsp-ui-sideline--diagnostics))
+          (lsp-ui-sideline--diagnostics bol eol))
         (when (and lsp-ui-sideline-show-code-actions (or (lsp--capability "codeActionProvider")
                                                          (lsp--registered-capability "textDocument/codeAction")))
           (lsp--send-request-async (lsp--make-request
@@ -394,9 +444,9 @@ to the language server."
                                     (if (equal lsp-ui-sideline-update-mode 'line)
                                         (list :textDocument doc-id
                                               :range (lsp--region-to-range bol eol)
-                                              :context (list :diagnostics (lsp--cur-line-diagnotics)))
+                                              :context (list :diagnostics (lsp-cur-line-diagnostics)))
                                       (lsp--text-document-code-action-params)))
-                                   #'lsp-ui-sideline--code-actions
+                                   (lambda (actions) (lsp-ui-sideline--code-actions actions bol eol))
                                    'alive))
         ;; Go through all symbols and request hover information.  Note that the symbols are
         ;; traversed backwards as `forward-symbol' with a positive argument will jump just past the
@@ -417,13 +467,22 @@ to the language server."
                 ;; Skip strings and comments
                 (when (and symbol (not in-string) outside-comment)
                   (push (list symbol tag bounds (lsp--position (1- line-widen) (- (point) bol))) symbols))))
-            (dolist (entry symbols)
-              (-let [(symbol tag bounds position) entry]
-                (lsp--send-request-async
-                 (lsp--make-request
-                  "textDocument/hover"
-                  (list :textDocument doc-id :position position))
-                 (lambda (info) (if info (lsp-ui-sideline--push-info symbol tag bounds info))))))))))))
+            (seq-do #'lsp--cancel-request lsp-ui-sideline--requests)
+            (dolist (ov lsp-ui-sideline--ovs)
+              (when (eq (overlay-get ov 'kind) 'info)
+                (setq lsp-ui-sideline--occupied-lines
+                  (delq (overlay-get ov 'position) lsp-ui-sideline--occupied-lines))
+                (delete-overlay ov)))
+            (setq lsp-ui-sideline--requests
+              (mapcar (lambda (it)
+                        (-let (((symbol tag bounds position) it))
+                          (plist-get (lsp-request-async
+                                       "textDocument/hover"
+                                       (list :textDocument doc-id :position position)
+                                       (lambda (info)
+                                         (when info (lsp-ui-sideline--push-info symbol tag bounds info bol eol))))
+                            :id)))
+                symbols))))))))
 
 (defun lsp-ui-sideline--stop-p ()
   "Return non-nil if the sideline should not be display."
@@ -439,7 +498,7 @@ COMMAND is `company-pseudo-tooltip-frontend' parameter."
     (setq lsp-ui-sideline--tag nil)))
 
 (defun lsp-ui-sideline ()
-  "Show informations of the current line."
+  "Show information for the current line."
   (if (lsp-ui-sideline--stop-p)
       (progn (setq lsp-ui-sideline--tag nil)
              (lsp-ui-sideline--delete-ov))
@@ -459,7 +518,7 @@ COMMAND is `company-pseudo-tooltip-frontend' parameter."
                                        (lsp-ui-sideline--run)))))))))
 
 (defun lsp-ui-sideline-toggle-symbols-info ()
-  "Toggle display of symbols informations.
+  "Toggle display of symbols information.
 This does not toggle display of flycheck diagnostics or code actions."
   (interactive)
   (when (bound-and-true-p lsp-ui-sideline-mode)
@@ -469,11 +528,21 @@ This does not toggle display of flycheck diagnostics or code actions."
 
 (defun lsp-ui-sideline--diagnostics-changed ()
   "Handler for flycheck notifications."
+  (lsp-ui-sideline--delete-ov)
   (setq lsp-ui-sideline--tag nil)
   (lsp-ui-sideline))
 
+(defun lsp-ui-sideline--erase (&rest _)
+  (when (bound-and-true-p lsp-ui-sideline-mode)
+    (ignore-errors
+      (lsp-ui-sideline--delete-ov)
+      (setq lsp-ui-sideline--tag nil))))
+
+(defvar lsp-ui-sideline-cmd-erase
+  '(kill-region))
+
 (define-minor-mode lsp-ui-sideline-mode
-  "Minor mode for showing information of current line."
+  "Minor mode for showing information for current line."
   :init-value nil
   :group lsp-ui-sideline
   (cond
@@ -481,20 +550,27 @@ This does not toggle display of flycheck diagnostics or code actions."
     (add-hook 'post-command-hook 'lsp-ui-sideline nil t)
     (advice-add 'company-pseudo-tooltip-frontend :before 'lsp-ui-sideline--hide-before-company)
     (add-hook 'lsp-after-diagnostics-hook 'lsp-ui-sideline--diagnostics-changed nil t)
+    (dolist (cmd lsp-ui-sideline-cmd-erase)
+      (advice-add cmd :before 'lsp-ui-sideline--erase))
     (when lsp-ui-sideline-show-diagnostics
       (setq-local flycheck-display-errors-function nil)))
    (t
     (setq lsp-ui-sideline--tag nil)
     (advice-remove 'company-pseudo-tooltip-frontend 'lsp-ui-sideline--hide-before-company)
     (lsp-ui-sideline--delete-ov)
-    (remove-hook 'lsp-after-diagnostics-hook 'lsp-ui-sideline--diagnostics-changed)
+    (remove-hook 'lsp-after-diagnostics-hook 'lsp-ui-sideline--diagnostics-changed t)
     (remove-hook 'post-command-hook 'lsp-ui-sideline t)
+    (dolist (cmd lsp-ui-sideline-cmd-erase)
+      (advice-remove cmd 'lsp-ui-sideline--erase))
     (when lsp-ui-sideline-show-diagnostics
       (kill-local-variable 'flycheck-display-errors-function)))))
 
 (defun lsp-ui-sideline-enable (enable)
   "Enable/disable `lsp-ui-sideline-mode'."
-  (lsp-ui-sideline-mode (if enable 1 -1)))
+  (lsp-ui-sideline-mode (if enable 1 -1))
+  (if enable
+      (add-hook 'before-revert-hook 'lsp-ui-sideline--delete-ov nil t)
+    (remove-hook 'before-revert-hook 'lsp-ui-sideline--delete-ov t)))
 
 (provide 'lsp-ui-sideline)
 ;;; lsp-ui-sideline.el ends here
